from enum import Enum
import numpy as np


class Coord(Enum):
    CARTESIAN = 1
    CYLINDRICAL = 2


class Mesh(object):

    def __init__(self, name=None, vertices=None, faces=None, coord=None):

        if name is not None and not isinstance(name, (str,)):
                raise TypeError('Name property must be a string!')

        if vertices is not None:

            if not isinstance(vertices, (np.ndarray,)):
                raise TypeError('Vertices must be a numpy array with shape (n, 3)')

            vert_shape = vertices.shape

            if len(vert_shape) != 2 or vert_shape[1] != 3:
                raise TypeError('Vertices must be a numpy array with shape (n, 3)')

        if faces is not None:

            if not isinstance(faces, (np.ndarray)):
                raise TypeError('Faces must be a numpy array with shape (n, 4)')

            face_shape = faces.shape

            if len(face_shape) != 2 or face_shape[1] != 4:
                raise TypeError('Faces must be a numpy array with shape (n, 4)')

        self._name = name
        self._vertices = vertices
        self._faces = faces
        self._coord = Coord.CARTESIAN if coord is None else coord

    @property
    def name(self):
        return "" if self._name is None else self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, (str,)):
            raise TypeError("Name property must be a string!")

        self._name = value

    @property
    def faces(self):
        return self._faces

    @property
    def num_faces(self):
        if self._faces is None:
            return 0
        else:
            return self._faces.shape[0]

    @property
    def vertices(self):
            return self._vertices

    @property
    def num_vertices(self):
        if self._vertices is None:
            return 0
        else:
            return self._vertices.shape[0]

    def __repr__(self):
        s = "Mesh:\t\t{}\n".format(self.name)
        s += "Vertices:\t{}\n".format(self.num_vertices)
        s += "Faces:\t\t{}\n".format(self.num_faces)
        return s

    def save(self, filename):
        with open(filename, 'w') as f:

            # Write the object's name
            f.write("o {}\n".format(self._name))

            # Write all the vertices
            for vert in self.vertices:
                f.write("v {} {} {}\n".format(*vert))

            # Finally all the faces
            for face in self.faces:
                face_str = "f " + " ".join(str(f) for f in face) + "\n"
                f.write(face_str)



import numpy as np
from .mesh import Coord


class VertexArray(object):


    def __init__(self, verts=None, coord=None):

        if verts is not None:

            if not isinstance(verts, (np.ndarray,)):
                raise TypeError('Vertices must be a numpy array with shape (n, 3)')

            shape = verts.shape

            if len(shape) != 2 or shape[1] != 3:
                raise TypeError('Vertices must be a numpy array with shape (n, 3)')

        if coord is not None:

            if not isinstance(coord, (Coord,)):
                message = "Coordinate System must be one of "
                message += ", ".join(c.name for c in iter(Coord))
                raise TypeError(message)

        self._data = verts
        self._coord = Coord.CARTESIAN if coord is None else coord



import numpy as np
from .mesh import Mesh


class Translate(object):
    """
    Translate an object by [dx, dy, dz]
    """

    def __init__(self, dx, dy, dz):

        self._offset = np.array([dx, dy, dz])

    def __rrshift__(self, mesh):

        vs = mesh._vertices
        vs += self._offset

        return Mesh(mesh.name, vertices=vs, faces=mesh.faces,
                    coord=mesh._coord)


class Scale(object):
    """
    Scale either in all directions or a specified axis
    """
    def __init__(self, all=1.):
        self._all = all

    def __rrshift__(self, mesh):

        vs = mesh._vertices
        vs *= self._all

        return Mesh(mesh.name, vertices=vs, faces=mesh.faces,
                    coord=mesh._coord)


def translate(dx=0., dy=0., dz=0.):
    return Translate(dx, dy, dz)


def scale(x):
    return Scale(all=x)



import numpy as np
import numpy.random as npr
from hypothesis.strategies import integers, floats, composite

# In this file we define a number of strategies representing
# particular data types. Defining them here in one place will
# help keep ourselves consistent.

# A "real" number in the range of +- 1 million
real = floats(min_value=-1e6, max_value=1e6)


# A "size" a non negative integer representing the length of arrays etc.
size = integers(min_value=1, max_value=512)


# A stategy representing an aribrary array of vertices in cartesian coordinates
@composite
def vertices(draw):
    num = draw(size)

    return npr.rand(num, 3)


# A strategy representing an arbitrary array of faces
@composite
def faces(draw):
    num = draw(size)
    return npr.randint(1, 256, size=(num, 4))



from pytest import raises
from hypothesis import given
from hypothesis.strategies import text
from .strategies import vertices, faces


import numpy as np
from topos import Mesh
from topos.mesh import Coord


class TestInit(object):

    def test_no_args(self):
        mesh = Mesh()

        assert mesh._name is None
        assert mesh._vertices is None
        assert mesh._faces is None
        assert mesh._coord == Coord.CARTESIAN

    @given(name=text())
    def test_name(self, name):
        mesh = Mesh(name=name)

        assert mesh._name == name

        with raises(TypeError) as err:
            mesh = Mesh(name=2)

        assert 'must be a string' in str(err.value)

    @given(vertices=vertices())
    def test_vertices(self, vertices):

        mesh = Mesh(vertices=vertices)
        assert (mesh._vertices == vertices).all()

    def test_vertices_bad_values(self):

        with raises(TypeError) as err:
            mesh = Mesh(vertices=[4, 5, 6])

        assert 'array with shape (n, 3)' in str(err.value)

        verts = np.array([[1, 2, 3, 4],
                          [5, 6, 7, 8]])

        with raises(TypeError) as err:
            mesh = Mesh(vertices=verts)

        assert 'array with shape (n, 3)' in str(err.value)

    @given(faces=faces())
    def test_faces(self, faces):
        mesh = Mesh(faces=faces)
        assert (mesh.faces == faces).all()

    def test_faces_bad_values(self):

        with raises(TypeError) as err:
            mesh = Mesh(faces=[2, 3, 4])

        assert 'array with shape (n, 4)' in str(err.value)

        faces = np.array([[[1, 2, 3, 4],
                            [2, 3, 4, 5]]])

        with raises(TypeError) as err:
            mesh = Mesh(faces=faces)

        assert 'array with shape (n, 4)' in str(err.value)



class TestProperties(object):


    @given(name=text())
    def test_name_property(self, name):
        mesh = Mesh()

        # By default no name should be the empty string
        assert mesh.name == ""

        # And of course setting the name should work
        mesh.name = name
        assert mesh.name == name

    def test_name_property_bad_value(self):
        mesh = Mesh("square")

        # Names should only be a string
        with raises(TypeError) as err:
            mesh.name = 2

        assert "must be a string" in str(err.value)

        # Name should be unchanged
        assert mesh.name == "square"



from .mesh import Mesh, Coord
from .generators import planar_faces, planar_vertices,\
        cylindrical_faces, cylindrical_vertices


class Plane(Mesh):

    def __init__(self, N=4, size=1.0):

        verts = size * planar_vertices(N)
        faces = planar_faces(N)
        super().__init__(name="Plane", vertices=verts, faces=faces)


class Cylinder(Mesh):

    def __init__(self, N_theta=8, N_z=6):

        verts = cylindrical_vertices(N_theta, N_z)
        faces = cylindrical_faces(N_theta, N_z)

        super().__init__(name="Cylinder", vertices=verts, faces=faces,
                         coord=Coord.CYLINDRICAL)



from setuptools import setup


def readme():
    with open("README.rst") as f:
        return f.read()


exec(open('topos/version.py', 'r').read())


setup(name='topos',
      version=__version__,
      description="Mesh generation with a blend of python and maths",
      long_description=readme(),
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3 :: Only',
          'Topic :: Multimedia :: Graphics'
      ],
      author='Alex Carney',
      author_email="alcarneyme@gmail.com",
      license='MIT',
      packages=['topos'],
      install_requires=[
          'numpy'
      ],
      setup_requires=['pytest-runner'],
      test_suite='tests',
      tests_require=['pytest', 'hypothesis'],
      python_requires='>=3.0',
      include_package_data=True,
      zip_safe=False)



import numpy as np
from math import pi


def planar_vertices(N):
    """
    Generate a grid of vertices in the X-Y Plane on
    :math:`[0, 1] \times [0, 1]` at the given resolution
    N
    """
    # Generate a grid of x-y values at the desired resolution
    xs = np.linspace(0, 1, N)
    XS, YS = np.meshgrid(xs, xs)

    # Reshape the grids to be one long array
    XS.shape = (N*N,)
    YS.shape = (N*N,)
    zeros = np.zeros(XS.shape)

    # Return the vertices all at z=0
    return np.dstack([XS, YS, zeros])


def cylindrical_vertices(N_theta, N_z):
    """
    Generate vertices in the shape of a cylinder
    coordinates are in cylindrical coordinates
    in the format (t, z, r)
    """

    ts = np.linspace(0, 2*pi, N_theta)
    zs = np.linspace(0, 1, N_z)

    TS, ZS = np.meshgrid(ts, zs)
    TS.shape = (N_theta * N_z,)
    ZS.shape = (N_theta * N_z,)
    ones = np.full(ZS.shape, 1.)

    return np.dstack([TS, ZS, ones])[0]


def cylindrical_faces(N_theta, N_z, close_loop=True):
    """
    Generate the face definitions for an object of cylindrical geometry
    optionally closing the loop
    """
    faces = []

    for j in range(0, N_z - 1):

            # Here I am using the fact that python promotes True/False to
            # 1/0 respectively when used in arithmetic to calculate the range
            for i in range(1, N_theta + (close_loop * 1)):

                # To get the normals pointing in the right direction we need
                # to define the corners in anti-clockwise order. So we will
                # # do it as follows:
                #      Lower left -> Lower right -> Upper right -> Upper left
                lower_left = j * N_theta + i
                lower_right = j*N_theta + i + 1
                upper_right = (j + 1) * N_theta + i + 1
                upper_left = (j + 1) * N_theta + i

                # Only the 'right hand' values could be problematic when closing
                # the loop
                if lower_right >= (j+1) * N_theta:
                    lower_right -= N_theta - 1

                if upper_right >= (j+2) * N_theta:
                    upper_right -= N_theta - 1

                # Add the face to the mesh
                faces.append((lower_left, lower_right, upper_right, upper_left))

    return np.array(faces)


def planar_faces(N):
    """
    Generate the face definitions for a planar grid of size N vertices
    by N vertices. To get the normals facing the right direction the elements
    of each face reference the vertex indices in the following order:

    4        <- 3
    *-----------*
    |           |
    |           |
    |           | ^
    |           | |
    *-----------* 2
    1 ->

    Parameters
    ----------
    N: int
        The number of vertices to include in the side of the mesh

    Returns
    -------
    faces: numpy.ndarray
        A numpy array of shape :math:`((N-1)^2, 4)` representing the faces
        of the mesh
    """

    # TODO: Is there a more "numpy" way to do this?
    faces = []
    for j in range(0, N - 1):
        for i in range(1, N):

            lower_left = i + (j * N)
            lower_right = (i + 1) + (j * N)
            upper_right = (i + 1) + ((j + 1) * N)
            upper_left = i + ((j + 1) * N)

            faces.append([lower_left, lower_right, upper_right, upper_left])

    return np.array(faces)



from pytest import raises
from hypothesis import given
import numpy as np


from topos.mesh import Coord
from topos.vertices import VertexArray


class TestInit(object):

    def test_no_args(self):

        verts = VertexArray()

        assert verts._data is None
        assert verts._coord == Coord.CARTESIAN

    def test_bad_verts(self):

        with raises(TypeError) as err:
            verts = VertexArray("String")

        assert 'array with shape (n, 3)' in str(err.value)

        with raises(TypeError) as err:
            verts = VertexArray(np.array([[2], [1], [3]]))

        assert 'array with shape (n, 3)' in str(err.value)

    def test_bad_coord(self):

        with raises(TypeError) as err:
            verts = VertexArray(coord="Polar")

        assert 'must be one of' in str(err.value)
